import groovy.transform.Field

// ビルドキャッシュ用のバケット
// https://console.cloud.google.com/storage/browser/magellan-ci-jenkins-cache?project=magellan-ci&organizationId=77670238700
@Field def bucket_name = 'magellan-ci-jenkins-cache'

def cache_dir(){
  // gcsfuse でマウントされた GCS バケットの、Jenkins のジョブごとに分けられたディレクトリ。
  // 各ジョブのキャッシュファイルはこのディレクトリ以下に置く。
  // キャッシュファイルのフルパスは gen_cache_file() で生成する。
  return "/gcs/${bucket_name}/jobs/${job_name()}/files"
}

def job_name(){
  if (env.BRANCH_NAME && env.JOB_BASE_NAME) {
    // Multi-Branch Pipeline の場合
    // JOB_NAME には "magellan-maneuvers/xxx/features%2Fjenkins-cache-gcs" のような
    // Jenkins の絶対プロジェクト名が入っているので、"magellan-maneuvers/xxx" を返します。
    return env.JOB_NAME - ('/' + env.JOB_BASE_NAME)
  } else {
    // Multi-Branch Pipeline 以外の場合
    // Jenkins の絶対プロジェクト名をそのまま返します。
    return env.JOB_NAME
  }
}

// キャッシュファイルのフルパスを返す
def gen_cache_file(Map params = [:]){
  def prefix   = params.prefix
  def lockfile = params.lockfile
  def suffix   = params.suffix   ?: 'tar.gz'

  if (lockfile){
    def checksum = sha1(lockfile)
    return "${cache_dir()}/${prefix}_${checksum}.${suffix}"
  } else {
    return "${cache_dir()}/${prefix}.${suffix}"
  }
}

def cache_exist(cache_file){
  // gcsfuse 経由でファイルを作成すると、ファイルが存在するが 0 バイト、という状況がありえるので、
  // -f ではなく -s (ファイルが存在し、かつ大きさが 0 より大きい) を使います。
  def ret = sh(
    script: "[ -s ${cache_file} ]",
    returnStatus: true
  )
  return ret == 0
}

// (汎用) キャッシュを展開する
def restore_cache(Map params = [:]){
  def prefix   = params.prefix
  def lockfile = params.lockfile
  def suffix   = params.suffix   ?: 'tar.gz'

  if (lockfile) {
    lockfile = lockfile_in_git(lockfile)
  }
  def cache_file = gen_cache_file(prefix: prefix, lockfile: lockfile, suffix: suffix)

  if (cache_exist(cache_file)){
    switch(suffix){
      case 'tar.zst':
        // TODO: tar コマンドが zstd をサポートすると不要になる。具体的には debian-10 buster では不要になる
        sh "zstdcat ${cache_file} | tar xf -"
        break
      default:
        sh "tar xf ${cache_file}"
    }
  }
}

// (汎用) キャッシュを作成する
def save_cache(Map params = [:]){
  def prefix   = params.prefix
  def lockfile = params.lockfile
  def targets  = params.targets
  def suffix   = params.suffix   ?: 'tar.gz'

  if (lockfile) {
    lockfile = lockfile_in_git(lockfile)
  }
  def cache_file = gen_cache_file(prefix: prefix, lockfile: lockfile, suffix: suffix)

  // キャッシュが存在しない場合だけでなく、lockfile が git リポジトリに含まれていない場合もキャッシュを作成（or更新）します。
  if (!lockfile || !cache_exist(cache_file)) {
    // cache_file は cache_dir 直下のファイルとは限らないので、cache_file の dirname に対して mkdir します。
    sh "mkdir -p \$(dirname ${cache_file})"
    switch(suffix){
      case 'tar.zst':
        // TODO: tar コマンドが zstd をサポートすると不要になる。具体的には debian-10 buster では不要になる
        sh "tar cf - ${targets.join(' ')} | zstd - -o ${cache_file} -f"
        break
      default:
        sh "tar cf ${cache_file} -a ${targets.join(' ')}"
    }
  }
}

// lockfile が git リポジトリ内にあるファイルならファイル名を返し、
// git リポジトリ内になければ null を返します。
def lockfile_in_git(lockfile){
  def git_ls_files = sh(
    script: "git ls-files ${lockfile}",
    returnStdout: true
  ).trim()
  return lockfile == git_ls_files ? lockfile : null
}

// bundler 用のキャッシュを展開する
def restore_bundler_cache(Map params = [:]){
  def prefix   = params.prefix   ?: 'bundler'
  def lockfile = params.lockfile ?: 'Gemfile.lock'
  def suffix   = params.suffix   ?: 'tar.gz'

  prefix = "${prefix}_ruby-${ruby_api_version()}"

  restore_cache(prefix: prefix, lockfile: lockfile, suffix: suffix)
}

// bundler 用のキャッシュを作成する
def save_bundler_cache(Map params = [:]){
  def prefix   = params.prefix   ?: 'bundler'
  def lockfile = params.lockfile ?: 'Gemfile.lock'
  def targets  = params.targets  ?: ['vendor/bundle']
  def suffix   = params.suffix   ?: 'tar.gz'

  prefix = "${prefix}_ruby-${ruby_api_version()}"

  save_cache(prefix: prefix, lockfile: lockfile, targets: targets, suffix: suffix)
}

// Ruby の API バージョンを返します。
def ruby_api_version(){
  if (is_in_docker_container()){
    // Docker コンテナ上の場合
    return sh(
      script: '''
        ruby -e "puts RbConfig::CONFIG[%(ruby_version)]"
      ''',
      returnStdout: true
    ).trim()
  } else {
    // VM 上の場合 (rbenv でインストールした ruby を使う)
    return sh(
      script: '''
        bash -lc 'ruby -e "puts RbConfig::CONFIG[%(ruby_version)]"'
      ''',
      returnStdout: true
    ).trim()
  }
}

// Docker コンテナ内かどうかを判定します。
def is_in_docker_container(){
  // 参考: https://github.com/chef/ohai/blob/15.1.0/lib/ohai/plugins/linux/virtualization.rb#L190-L228
  def ret = sh(
    script: '[ -f /.dockerenv -o -f /.dockerinit ]',
    returnStatus: true
  )
  return ret == 0
}

return this
